@@ -1,3 +1,14 @@
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TupleSections #-}
+
+#if __GLASGOW_HASKELL__ >= 800
+{-# OPTIONS_GHC -fno-warn-unused-top-binds #-}
+#endif
+
 -- | Agda backend.
 --
 -- Generate bindings to Haskell data types for use in Agda.
@@ -108,17 +119,6 @@
 -- > {-# COMPILE GHC printDef     = \ d -> Data.Text.pack (printTree (d :: CPP.Abs.Def))  #-}
 -- > {-# COMPILE GHC printProgram = \ p -> Data.Text.pack (printTree (p :: CPP.Abs.Program)) #-}
 
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE LambdaCase #-}
-{-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE NoImplicitPrelude #-}
-{-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections #-}
-
-#if __GLASGOW_HASKELL__ >= 800
-{-# OPTIONS_GHC -fno-warn-unused-top-binds #-}
-#endif
-
 module BNFC.Backend.Agda (makeAgda) where
 
 import Prelude hiding ((<>))
@@ -143,8 +143,6 @@ import BNFC.Options                (SharedOptions, TokenText(..), tokenText)
 import BNFC.PrettyPrint
 import BNFC.Utils                  (ModuleName, replace, when, table)
 
-type List1 = List1.NonEmpty
-
 -- | How to print the types of constructors in Agda?
 
 data ConstructorStyle
@@ -331,25 +329,32 @@ imports
   -> Doc
 imports numeric layout pos = vcat . map prettyImport . concat $
   [ when layout
-    [ ("Agda.Builtin.Bool",   [("Bool", boolT)]) ]
-  , [ ("Agda.Builtin.Char",   [("Char", charT)]) ]
+    [ ("Agda.Builtin.Bool",   [],            [("Bool", boolT)]) ]
+  , [ ("Agda.Builtin.Char",   [charT],       []               ) ]
   , when (numeric == YesImportNumeric) importNumeric
-  , [ ("Agda.Builtin.List",   [("List", listT)]) ]
+  , [ ("Agda.Builtin.List",   ["[]", "_∷_"], [("List", listT)]) ]
   , when pos
-    [ ("Agda.Builtin.Nat",    [("Nat" , natT )]) ]
-  , [ ("Agda.Builtin.String", [("String", stringT), ("primStringFromList", stringFromListT) ]) ]
+    [ ("Agda.Builtin.Nat",    [],            [("Nat" , natT )]) ]
+  , [ ("Agda.Builtin.String", [], [("String", stringT), ("primStringFromList", stringFromListT) ]) ]
   ]
   where
-  importNumeric :: [(String, [(String, Doc)])]
+  importNumeric :: [(String, [Doc], [(String, Doc)])]
   importNumeric =
-    [ ("Agda.Builtin.Float public", [("Float", doubleT)])
-    , ("Agda.Builtin.Int   public", [("Int", integerT)])
+    [ ("Agda.Builtin.Float public", [], [("Float", doubleT)])
+    , ("Agda.Builtin.Int   public", [], [("Int", integerT)])
+    , ("Agda.Builtin.Int"         , [], [("pos", "#pos")])
     ]
-  prettyImport :: (String, [(String, Doc)]) -> Doc
-  prettyImport (m, ren) = prettyList 2 pre lparen rparen semi $
-    map (\ (x, d) -> hsep [text x, "to", d ]) ren
+  prettyImport :: (String, [Doc], [(String, Doc)]) -> Doc
+  prettyImport (m, use, ren)
+    | null ren  = pre
+    | otherwise = prettyList 2 pre lparen rparen semi $
+        map (\ (x, d) -> hsep [text x, "to", d ]) ren
     where
-    pre = hsep [ "open", "import", text m, "using", "()", "renaming" ]
+    pre = hsep $ concat
+      [ [ "open", "import", text m ]
+      , [ "using", parens $ hcat $ punctuate "; " use ]
+      , [ "renaming" | not (null ren) ]
+      ]
 
 -- | Import Agda AST.
 --
@@ -462,7 +467,7 @@ absyn  amod  style ds = vsep . ("mutual" :) . concatMap (map (nest 2) . prData a
 --
 prData :: ModuleName -> ConstructorStyle -> Data -> [Doc]
 prData amod style (Cat d, cs) = prData' amod style d cs
-prData _    _     (c    , _ ) = error $ "prData: unexpected category " ++ show c
+prData _    _     (c    , _ ) = error $ "prData: unexpected category " ++ prettyShow c
 
 -- | Pretty-print Agda data types and pragmas.
 --
@@ -649,23 +654,24 @@ incr = Map.alter $ maybe (Just 1) (Just . succ)
 
 -- | Generate Haskell code for the @define@d constructors.
 definedRules :: CF -> Doc
-definedRules cf = vsep [ mkDef f xs e | FunDef f xs e <- cfgPragmas cf ]
+definedRules cf = vsep $ map mkDef $ definitions cf
   where
-    mkDef f xs e = vcat $ map text $ concat
-      [ [ unwords [ mkDefName f, ":", typeToHaskell' "→" $ wpThing t ]
+    mkDef (Define f args e t) = vcat $ concat
+      [ [ text $ unwords [ mkDefName f, ":", typeToHaskell' "→" $ wpThing t ]
         | t <- maybeToList $ sigLookup f cf
         ]
-      , [ unwords $ concat
-          [ [ mkDefName f, "=", "λ" ]
-          , map agdaLower xs
-          , [ "→", show $ sanitize e ]
+      , [ hsep $ concat
+          [ [ text (mkDefName f), "=", "λ" ]
+          , map (text . agdaLower . fst) args
+          , [ "→", pretty $ sanitize e ]
           ]
         ]
       ]
     sanitize = \case
-      App x es      -> App (agdaLower x) $ map sanitize es
+      App "(:)" t es-> App "_∷_" t $ map sanitize es
+      App x t es    -> App (agdaLower x) t $ map sanitize es
       Var x         -> Var $ agdaLower x
-      e@LitInt{}    -> e
+      e@LitInt{}    -> App "#pos" dummyType [e]
       e@LitDouble{} -> e
       e@LitChar{}   -> e
       e@LitString{} -> e
