@@ -9,6 +9,7 @@
 module BNFC.Backend.Haskell (makeHaskell, AlexVersion(..), makefile, testfile) where
 
 import qualified Control.Monad as Ctrl
+import Data.Maybe      (isJust)
 import System.FilePath ((<.>), (</>), pathSeparator)
 import Text.Printf     (printf)
 import Text.PrettyPrint
@@ -58,7 +59,7 @@ makeHaskell opts cf = do
         liftIO $ printf "Use Alex 3 to compile %s.\n" (alexFile opts)
 
     Ctrl.when (hasLayout cf) $ mkfile (layoutFile opts) $
-      cf2Layout (tokenText opts) layMod lexMod cf
+      cf2Layout layMod lexMod cf
 
     -- Generate Happy parser and matching test program.
     do
@@ -84,7 +85,7 @@ makeHaskell opts cf = do
     Ctrl.when (agda opts) $ makeAgda time opts cf
 
     -- Generate Makefile.
-    Makefile.mkMakefile opts $ makefile opts
+    Makefile.mkMakefile opts $ makefile opts cf
 
 
 -- | Generate the makefile (old version, with just one "all" target).
@@ -186,9 +187,10 @@ makefileHeader Options{ glr } = vcat
 -- | Generate the makefile.
 makefile
   :: Options
+  -> CF
   -> String    -- ^ Filename of the makefile.
   -> Doc       -- ^ Content of the makefile.
-makefile opts makeFile = vcat
+makefile opts cf makeFile = vcat
   [ makefileHeader opts
   , phonyRule
   , defaultRule
@@ -234,11 +236,16 @@ makefile opts makeFile = vcat
     where
     recipe    = unwords [ "bnfc", printOptions opts{ make = Nothing } ]
     tgts      = unwords . map ($ opts) . concat $
-      [ [ absFile, alexFile, happyFile, printerFile, tFile ]
+      [ [ absFile ]
+      , [ layoutFile | lay ]
+      , [ alexFile, happyFile, printerFile, tFile ]
       , when (agda opts)
         [ agdaASTFile, agdaParserFile, agdaLibFile, agdaMainFile ]
       ]
 
+  lay :: Bool
+  lay = hasLayout cf
+
   -- | Rule to invoke @happy@.
   happyRule :: Doc
   happyRule = Makefile.mkRule "%.hs" [ "%.y" ] [ "${HAPPY} ${HAPPY_OPTS} $<" ]
@@ -254,28 +261,33 @@ makefile opts makeFile = vcat
     tgt :: String
     tgt = tFileExe opts
     deps :: [String]
-    deps = map ($ opts)
-      [ absFile
-      , alexFileHs
-      , happyFileHs
-      , printerFile
-      , tFile
+    deps = map ($ opts) $ concat
+      [ [ absFile ]
+      , [ layoutFile | lay ]
+      , [ alexFileHs
+        , happyFileHs
+        , printerFile
+        , tFile
+        ]
       ]
 
   -- | Rule to build Agda parser.
   agdaRule :: Doc
   agdaRule = Makefile.mkRule "Main" deps [ "agda --no-libraries --ghc --ghc-flag=-Wwarn $<" ]
     where
-    deps = map ($ opts)
-      [ agdaMainFile  -- must be first!
-      , agdaASTFile
-      , agdaParserFile
-      , agdaLibFile
-      -- Haskell modules bound by Agda modules:
-      , errFile
-      , alexFileHs
-      , happyFileHs
-      , printerFile
+    deps = map ($ opts) $ concat
+      [ [ agdaMainFile  -- must be first!
+        , agdaASTFile
+        , agdaParserFile
+        , agdaLibFile
+        -- Haskell modules bound by Agda modules:
+        , errFile
+        ]
+      , [ layoutFile | lay ]
+      , [ alexFileHs
+        , happyFileHs
+        , printerFile
+        ]
       ]
 
 testfile :: Options -> CF -> String
@@ -285,18 +297,18 @@ testfile opts cf = unlines $ concat $
     , "module Main where"
     , ""
     , "import Prelude"
-    , "  ( ($)"
+    , "  ( ($), (.)"
     ]
   , [ "  , Bool(..)" | lay ]
   , [ "  , Either(..)"
     , "  , Int, (>)"
-    , "  , String, (++), unlines"
+    , "  , String, (++), concat, unlines"
     , "  , Show, show"
     , "  , IO, (>>), (>>=), mapM_, putStrLn"
     , "  , FilePath"
     ]
   , [ "  , getContents, readFile" | tokenText opts == StringToken ]
-  , [ "  , (.), error, flip, map, replicate, sequence_, zip" | use_glr ]
+  , [ "  , error, flip, map, replicate, sequence_, zip" | use_glr ]
   , [ "  )" ]
   , case tokenText opts of
       StringToken -> []
@@ -316,7 +328,7 @@ testfile opts cf = unlines $ concat $
   , table "" $ concat
     [ [ [ "import " , absFileM      opts , " (" ++ if_glr impTopCat ++ ")" ] ]
     , [ [ "import " , layoutFileM   opts , " ( resolveLayout )"      ] | lay     ]
-    , [ [ "import " , alexFileM     opts , " ( Token )"                          ]
+    , [ [ "import " , alexFileM     opts , " ( Token, mkPosToken )"              ]
       , [ "import " , happyFileM    opts , " ( " ++ impParser ++ ", myLexer" ++ impParGLR ++ " )" ]
       , [ "import " , printerFileM  opts , " ( Print, printTree )"               ]
       , [ "import " , templateFileM opts , " ()"                                 ]
@@ -367,14 +379,14 @@ testfile opts cf = unlines $ concat $
     , ""
     ]
   , if_glr $
-    [ "the_parser :: ParseFun " ++ show topType
+    [ "the_parser :: ParseFun " ++ catToStr topType
     , "the_parser = lift_parser " ++ render (parserName topType)
     , ""
     , liftParser
     ]
   ]
   where
-    lay         = hasLayout cf
+    lay         = isJust hasTopLevelLayout || not (null layoutKeywords)
     use_xml     = xml opts > 0
     xpr         = if use_xml then "XPrint a, "     else ""
     use_glr     = glr opts == GLR
@@ -386,8 +398,10 @@ testfile opts cf = unlines $ concat $
     impTopCat   = unwords [ "", identCat topType, "" ]
     impParGLR   = if_glr ", GLRResult(..), Branch, ForestId, TreeDecode(..), decode"
     myLLexer atom
-      | lay     = unwords [ "resolveLayout True $ myLexer", atom]
+      | lay     = unwords [ "resolveLayout", show useTopLevelLayout, "$ myLexer", atom]
       | True    = unwords [ "myLexer", atom]
+    (hasTopLevelLayout, layoutKeywords, _) = layoutPragmas cf
+    useTopLevelLayout = isJust hasTopLevelLayout
 
 
 runStd xml myLLexer = unlines $ concat
@@ -396,7 +410,8 @@ runStd xml myLLexer = unlines $ concat
    , "    Left err -> do"
    , "      putStrLn \"\\nParse              Failed...\\n\""
    , "      putStrV v \"Tokens:\""
-   , "      putStrV v $ show ts"
+   , "      mapM_ (putStrV v . showPosToken . mkPosToken) ts"
+   -- , "      putStrV v $ show ts"
    , "      putStrLn err"
    , "      exitFailure"
    , "    Right tree -> do"
@@ -407,6 +422,7 @@ runStd xml myLLexer = unlines $ concat
  , [ "      exitSuccess"
    , "  where"
    , "  ts = " ++ myLLexer "s"
+   , "  showPosToken ((l,c),t) = concat [ show l, \":\", show c, \"\\t\", show t ]"
    ]
  ]
 
