@@ -1,6 +1,7 @@
 {-# LANGUAGE DeriveTraversable #-}  -- implies DeriveFunctor, DeriveFoldable
 {-# LANGUAGE FlexibleInstances #-}  -- implies TypeSynonymInstances
 {-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE PatternSynonyms #-}
 {-# LANGUAGE RecordWildCards #-}
@@ -8,96 +9,21 @@
 {-
     BNF Converter: Abstract syntax
     Copyright (C) 2004  Author:  Markus Forsberg, Michael Pellauer, Aarne Ranta
-
+    Copyright (C) 2017-2021 Andreas Abel
 -}
 
-module BNFC.CF (
-            -- Types.
-            CF,
-            CFG(..),
-            Rule, Rul(..), npRule, valCat, lookupRule, InternalRule(..),
-            Pragma(..),
-            Exp(..),
-            Base(..), Type(..), Signature,
-            Literal,
-            Symbol,
-            KeyWord,
-            Position(..), noPosition, prettyPosition, npIdentifier,
-            WithPosition(..), blendInPosition,
-            RString, RCat,
-            Cat(..), strToCat, catToStr,
-            BaseCat, TokenCat,
-            catString, catInteger, catDouble, catChar, catIdent,
-            NonTerminal, SentForm,
-            Fun, RFun, IsFun(..),
-            Data,           -- describes the abstract syntax of a grammar
-            cf2data,        -- translates a grammar to a Data object.
-            cf2dataLists,   -- translates to a Data with List categories included.
-            getAbstractSyntax,
-            -- Literal categories, constants,
-            firstEntry,     -- the first entry or the first value category
-            baseTokenCatNames,  -- "Char", "Double", "Integer", "String"
-            specialCats,    -- ident
-            specialCatsP,   -- all literals
-            specialData,    -- special data
-            isCoercion,     -- wildcards in grammar (avoid syntactic clutter)
-            isDefinedRule,  -- defined rules (allows syntactic sugar)
-            isProperLabel,  -- not coercion or defined rule
-            allCats,        -- all categories of a grammar
-            allParserCats, allParserCatsNorm,
-            reallyAllCats,
-            allCatsNorm,
-            allCatsIdNorm,
-            allEntryPoints,
-            reservedWords,
-            cfTokens,
-            literals,
-            findAllReversibleCats, -- find all reversible categories
-            identCat,       -- transforms '[C]' to ListC (others, unchanged).
-            isParsable,
-            rulesForCat,    -- rules for a given category
-            rulesForNormalizedCat,    -- rules for a given category
-            ruleGroups,     -- Categories are grouped with their rules.
-            ruleGroupsInternals, --As above, but includes internal cats.
-            allNames,        -- Checking for non-unique names, like @Name. Name ::= Ident;@.
-            filterNonUnique,
-            isList,         -- Checks if a category is a list category.
-            isTokenCat, maybeTokenCat,
-            baseCat,
-            sameCat,
-            -- Information functions for list functions.
-            isNilFun,       -- empty list function? ([])
-            isOneFun,       -- one element list function? (:[])
-            hasOneFunc,
-            getCons,
-            getSeparatorByPrecedence,
-            isConsFun,      -- constructor function? (:)
-            isNilCons,      -- either three of above?
-            isEmptyListCat, -- checks if the list permits []
-            revSepListRule, -- reverse a rule, if it is of form C t [C].
-            normCat,
-            isDataCat, isDataOrListCat,
-            normCatOfList,  -- Removes precendence information and enclosed List. C1 => C, C2 => C
-            catOfList,
-            comments,       -- translates the pragmas into two list containing the s./m. comments
-            numberOfBlockCommentForms,
-            tokenPragmas,   -- get the user-defined regular expression tokens
-            tokenNames,     -- get the names of all user-defined tokens
-            precCat,        -- get the precendence level of a Cat C1 => 1, C => 0
-            precRule,       -- get the precendence level of the value category of a rule.
-            isUsedCat,
-            isPositionCat,
-            hasPositionTokens,
-            hasIdent, hasIdentLikeTokens,
-            hasLayout,
-            layoutPragmas,
-            sigLookup      -- Get the type of a rule label.
-           ) where
+module BNFC.CF where
+
+import Prelude hiding ((<>))
+
+import Control.Arrow ( (&&&) )
+import Control.Monad ( guard )
 
-import Control.Monad (guard)
 import Data.Char
-import Data.Function (on)
-import Data.List (nub, intersperse, sort, group, intercalate, find)
+import Data.Ord      ( Down(..) )
+import qualified Data.Either as Either
+import Data.Function ( on )
+import Data.List     ( nub, intersperse, sort, group, intercalate, find )
 import qualified Data.List as List
 import Data.List.NonEmpty (pattern (:|), (<|))
 import qualified Data.List.NonEmpty as List1
@@ -113,7 +39,8 @@ import BNFC.Par (pCat)
 import BNFC.Lex (tokens)
 import qualified BNFC.Abs as Abs
 
-import BNFC.Utils (spanEnd)
+import BNFC.PrettyPrint
+import BNFC.Utils (headWithDefault, spanEnd)
 
 type List1 = List1.NonEmpty
 
@@ -154,10 +81,10 @@ data InternalRule
   | Parsable  -- ^ ordinary rule (also for parser)
   deriving (Eq)
 
-instance (Show function) => Show (Rul function) where
-  show (Rule f cat rhs internal) = unwords $
-    (if internal == Internal then ("internal" :) else id) $
-    show f : "." : show cat : "::=" : map (either show id) rhs
+instance Pretty function => Pretty (Rul function) where
+  pretty (Rule f cat rhs internal) =
+    (if internal == Internal then ("internal" <+>) else id) $
+    pretty f <> "." <+> pretty cat <+> "::=" <+> sep (map (either pretty (text . show)) rhs)
 
 -- | A sentential form is a sequence of non-terminals or terminals.
 type SentForm = [Either Cat String]
@@ -178,8 +105,8 @@ data CFG function = CFG
     } deriving (Functor)
 
 
-instance (Show function) => Show (CFG function) where
-  show CFG{..} = unlines $ map show cfgRules
+-- instance (Show function) => Show (CFG function) where
+--   show CFG{..} = unlines $ map show cfgRules
 
 -- | Types of the rule labels, together with the position of the rule label.
 type Signature = Map String (WithPosition Type)
@@ -193,6 +120,14 @@ data Base = BaseT String
 data Type = FunT [Base] Base
     deriving (Eq, Ord)
 
+-- | Placeholder for a type.
+dummyBase :: Base
+dummyBase = BaseT "<DUMMY>"
+
+-- | Placeholder for a function type.
+dummyType :: Type
+dummyType = FunT [] dummyBase
+
 instance Show Base where
     show (BaseT x) = x
     show (ListT t) = "[" ++ show t ++ "]"
@@ -202,43 +137,38 @@ instance Show Type where
 
 -- | Expressions for function definitions.
 
-data Exp
-  = App String [Exp]  -- ^ (Possibly defined) label applied to expressions.
-  | Var String        -- ^ Function parameter.
-  | LitInt Integer
+data Exp' f
+  = App f Type [Exp' f]     -- ^ (Possibly defined) label applied to expressions.
+                            --   The function 'Type' is inferred by the type checker.
+  | Var       String        -- ^ Function parameter.
+  | LitInt    Integer
   | LitDouble Double
-  | LitChar Char
+  | LitChar   Char
   | LitString String
   deriving (Eq)
 
-instance Show Exp where
-    showsPrec p e =
-        case listView e of
-            Right es    ->
-                showString "["
-                . foldr (.) id (intersperse (showString ", ") $ map shows es)
-                . showString "]"
-            Left (Var x)    -> showString x
-            Left (App x []) -> showString x
-            Left (App  "(:)" [e1,e2]) ->
-                showParen (p>0)
-                $ showsPrec 1 e1
-                . showString " : "
-                . shows e2
-            Left (App x es) ->
-                showParen (p>1)
-                $ foldr (.) id
-                $ intersperse (showString " ")
-                $ showString x : map (showsPrec 2) es
-            Left (LitInt n)     -> shows n
-            Left (LitDouble x)  -> shows x
-            Left (LitChar c)    -> shows c
-            Left (LitString s)  -> shows s
-        where
-            listView (App "[]" []) = Right []
-            listView (App "(:)" [e1,e2])
-                | Right es <- listView e2   = Right $ e1:es
-            listView x = Left x
+type Exp = Exp' String
+
+instance (IsFun f, Pretty f) => Pretty (Exp' f) where
+  prettyPrec p e =
+    case listView e of
+      Right es           -> brackets $ hcat $ punctuate ", " $ map (prettyPrec 0) es
+      Left (Var x)       -> text x
+      Left (App f _ [])  -> prettyPrec p f
+      Left (App f _ [e1,e2])
+        | isConsFun f    -> parensIf (p > 0) $ hsep [ prettyPrec 1 e1, ":", prettyPrec 0 e2 ]
+      Left (App f _ es)  -> parensIf (p > 1) $ hsep $ prettyPrec 1 f : map (prettyPrec 2) es
+      Left (LitInt n)    -> (text . show) n
+      Left (LitDouble x) -> (text . show) x
+      Left (LitChar c)   -> (text . show) c
+      Left (LitString s) -> (text . show) s
+    where
+      listView (App f _ [])
+        | isNilFun f              = Right []
+      listView (App f _ [e1,e2])
+        | isConsFun f
+        , Right es <- listView e2 = Right $ e1:es
+      listView e0                 = Left e0
 
 -- | Pragmas.
 
@@ -247,11 +177,44 @@ data Pragma
   | CommentM (String, String)     -- ^  for multiple-line comments.
   | TokenReg RString Bool Reg     -- ^ for tokens
   | EntryPoints [RCat]
-  | Layout [String]
-  | LayoutStop [String]
-  | LayoutTop
-  | FunDef RFun [String] Exp
-  deriving (Show)
+  | Layout LayoutKeyWords
+  | LayoutStop [KeyWord]
+  | LayoutTop Symbol              -- ^ Separator for top-level layout.
+  | FunDef Define
+
+data Define = Define
+  { defName :: RFun
+  , defArgs :: Telescope  -- ^ Argument types inferred by the type checker.
+  , defBody :: Exp
+  , defType :: Base       -- ^ Type of the body, inferred by the type checker.
+  }
+
+-- | Function arguments with type.
+type Telescope = [(String, Base)]
+
+-- | For use with 'partitionEithers'.
+isFunDef :: Pragma -> Either Pragma Define
+isFunDef = \case
+  FunDef d -> Right d
+  p        -> Left  p
+
+-- | All 'define' pragmas of the grammar.
+definitions :: CFG f -> [Define]
+definitions cf = [ def | FunDef def <- cfgPragmas cf ]
+
+------------------------------------------------------------------------------
+-- Layout
+------------------------------------------------------------------------------
+
+type LayoutKeyWords = [(KeyWord, Delimiters)]
+
+-- | List delimiters.
+
+data Delimiters = Delimiters
+  { listSep   :: Symbol           -- ^ List separator.
+  , listOpen  :: Symbol           -- ^ List opening delimiter.
+  , listClose :: Symbol           -- ^ List closing delimiter.
+  } deriving Show
 
 -- | User-defined regular expression tokens
 tokenPragmas :: CFG f -> [(TokenCat,Reg)]
@@ -261,16 +224,21 @@ tokenPragmas cf = [ (wpThing name, e) | TokenReg name _ e <- cfgPragmas cf ]
 tokenNames :: CFG f -> [String]
 tokenNames cf = map fst (tokenPragmas cf)
 
-layoutPragmas :: CF -> (Bool,[String],[String])
-layoutPragmas cf = let ps = cfgPragmas cf in (
-  not (null [() | LayoutTop  <- ps]),   -- if there's layout betw top-level
-  concat [ss | Layout ss     <- ps],    -- layout-block starting words
-  concat [ss | LayoutStop ss <- ps]     -- layout-block ending words
+layoutPragmas :: CF -> (Maybe Symbol, LayoutKeyWords, [KeyWord])
+layoutPragmas cf =
+  ( listToMaybe [ sep | LayoutTop  sep <- ps ]   -- if there's top-level layout
+  , concat      [ kws | Layout     kws <- ps ]   -- layout-block inducing words
+  , concat      [ kws | LayoutStop kws <- ps ]   -- layout-block aborting words
   )
+  where
+  ps = cfgPragmas cf
+
+hasLayout_ :: (Maybe Symbol, LayoutKeyWords, [KeyWord]) -> Bool
+hasLayout_ (top, kws, _) = isJust top || not (null kws)   -- (True,[],_) means: top-level layout only
 
 hasLayout :: CF -> Bool
-hasLayout cf = case layoutPragmas cf of
-  (t,ws,_) -> t || not (null ws)   -- (True,[],_) means: top-level layout only
+hasLayout = hasLayout_ . layoutPragmas
+
 
 -- | Literal: builtin-token types Char, String, Ident, Integer, Double.
 type Literal = String
@@ -304,6 +272,8 @@ data WithPosition a = WithPosition
 instance Eq  a => Eq  (WithPosition a) where (==)    = (==)    `on` wpThing
 instance Ord a => Ord (WithPosition a) where compare = compare `on` wpThing
 
+instance Pretty a => Pretty (WithPosition a) where pretty = pretty . wpThing
+
 noPosition :: a -> WithPosition a
 noPosition = WithPosition NoPosition
 
@@ -361,9 +331,18 @@ catToStr = \case
   ListCat c    -> "[" ++ catToStr c ++ "]"
   CoercCat s i -> s ++ show i
 
+-- | This instance is for the Hspec test suite.
 instance Show Cat where
   show = catToStr
 
+instance Pretty Cat where
+  pretty = \case
+    Cat s        -> text s
+    TokenCat s   -> text s
+    ListCat c    -> brackets $ pretty c
+    CoercCat s i -> text s <> pretty i
+
+
 -- | Reads a string into a category. This should only need to handle
 -- the case of simple categories (with or without coercion) since list
 -- categories are parsed in the grammar already. To be on the safe side here,
@@ -395,6 +374,7 @@ catIdent   = "Ident"
 baseTokenCatNames :: [TokenCat]
 baseTokenCatNames = [ catChar, catDouble, catInteger, catString ]
 
+-- all literals
 -- the parser needs these
 specialCatsP :: [TokenCat]
 specialCatsP = catIdent : baseTokenCatNames
@@ -417,10 +397,7 @@ isDataOrListCat _               = True
 -- True
 
 sameCat :: Cat -> Cat -> Bool
-sameCat (CoercCat c1 _) (CoercCat c2 _) = c1 == c2
-sameCat (Cat c1)        (CoercCat c2 _) = c1 == c2
-sameCat (CoercCat c1 _) (Cat c2)        = c1 == c2
-sameCat c1              c2              = c1 == c2
+sameCat = (==) `on` normCat
 
 -- | Removes precedence information. C1 => C, [C2] => [C]
 normCat :: Cat -> Cat
@@ -438,6 +415,19 @@ identCat :: Cat -> String
 identCat (ListCat c) = "List" ++ identCat c
 identCat c = catToStr c
 
+identType :: Base -> String
+identType (ListT t) = "List" ++ identType t
+identType (BaseT s) = s
+
+-- | Reconstruct (non-coercion) category from a type, given a list of
+-- what should be the token categories.
+catOfType :: [TokenCat] -> Base -> Cat
+catOfType tk = \case
+  ListT t -> ListCat $ catOfType tk t
+  BaseT s
+   | s `elem` tk -> TokenCat s
+   | otherwise   -> Cat s
+
 isList :: Cat -> Bool
 isList (ListCat _) = True
 isList _           = False
@@ -479,38 +469,46 @@ instance IsString RFun where
 class IsFun a where
   funName :: a -> String
 
+  isNilFun    :: a -> Bool   -- ^ Is this the constructor for empty lists?
+  isOneFun    :: a -> Bool   -- ^ Is this the constructor for singleton lists?
+  isConsFun   :: a -> Bool   -- ^ Is this the list constructor?
+  isConcatFun :: a -> Bool   -- ^ Is this list concatenation?
+
+  -- | Is this function just a coercion? (I.e. the identity)
+  isCoercion :: a -> Bool
+
+  isNilFun    = funNameSatisfies (== "[]")
+  isOneFun    = funNameSatisfies (== "(:[])")
+  isConsFun   = funNameSatisfies (== "(:)")
+  isConcatFun = funNameSatisfies (== "(++)")
+  isCoercion  = funNameSatisfies (== "_")
+
 instance IsFun String where
   funName = id
 
 instance IsFun a => IsFun (WithPosition a) where
   funName = funName . wpThing
 
+instance IsFun a => IsFun (Rul a) where
+  funName = funName . funRule
+
+instance IsFun a => IsFun (k, a) where
+  funName = funName . snd
+
 funNameSatisfies :: IsFun a => (String -> Bool) -> a -> Bool
 funNameSatisfies f = f . funName
 
--- | Is this function just a coercion? (I.e. the identity)
-isCoercion :: IsFun a => a -> Bool
-isCoercion = funNameSatisfies (== "_") -- perhaps this should be changed to "id"?
-
 isDefinedRule :: IsFun a => a -> Bool
 isDefinedRule = funNameSatisfies $ \case
   (x:_) -> isLower x
   []    -> error "isDefinedRule: empty function name"
 
--- isDefinedRule :: Fun -> Bool
--- isDefinedRule (WithPosition _ (x:_)) = isLower x
--- isDefinedRule (WithPosition pos []) = error $
---   unwords [ prettyPosition pos, "isDefinedRule: empty function name" ]
-
+-- not coercion or defined rule
 isProperLabel :: IsFun a => a -> Bool
 isProperLabel f = not (isCoercion f || isDefinedRule f)
 
-isNilFun, isOneFun, isConsFun, isNilCons,isConcatFun :: IsFun a => a -> Bool
+isNilCons :: IsFun a => a -> Bool
 isNilCons f = isNilFun f || isOneFun f || isConsFun f || isConcatFun f
-isNilFun    = funNameSatisfies (== "[]")
-isOneFun    = funNameSatisfies (== "(:[])")
-isConsFun   = funNameSatisfies (== "(:)")
-isConcatFun = funNameSatisfies (== "(++)")
 
 ------------------------------------------------------------------------------
 
@@ -688,14 +686,16 @@ cf2data' predicate cf =
   [(cat, nub (map mkData [r | r <- cfgRules cf,
                               let f = funRule r,
                               not (isDefinedRule f),
-                              not (isCoercion f), sameCat cat (valCat r)]))
+                              not (isCoercion f), cat == normCat (valCat r)]))
       | cat <- nub $ map normCat $ filter predicate $ reallyAllCats cf ]
- where
+  where
   mkData (Rule f _ its _) = (wpThing f, [normCat c | Left c <- its ])
 
+-- translates a grammar to a Data object.
 cf2data :: CF -> [Data]
 cf2data = cf2data' $ isDataCat . normCat
 
+-- translates to a Data with List categories included.
 cf2dataLists :: CF -> [Data]
 cf2dataLists = cf2data' $ isDataOrListCat . normCat
 
@@ -711,35 +711,24 @@ sigLookup f = Map.lookup (funName f) . cfgSignature
 isParsable :: Rul f -> Bool
 isParsable = (Parsable ==) . internal
 
+hasNilRule :: [Rule] -> Maybe Rule
+hasNilRule = List.find isNilFun
 
--- | Checks if the list has a non-empty rule.
-hasOneFunc :: [Rule] -> Bool
-hasOneFunc = any (isOneFun . funRule)
+-- | Gets the singleton rule out of the rules for a list.
+hasSingletonRule :: [Rule] -> Maybe Rule
+hasSingletonRule = List.find isOneFun
 
--- | Gets the separator for a list.
-getCons :: [Rule] -> String
-getCons rs = case find (isConsFun . funRule) rs of
-    Just (Rule _ _ cats _) -> seper cats
-    Nothing              -> error $ "getCons: no construction function found in "
-                                  ++ intercalate ", " (map (show . funRule) rs)
-  where
-    seper [] = []
-    seper (Right x:_) = x
-    seper (Left _:xs) = seper xs
+-- | Sort rules by descending precedence.
 
--- | Helper function that gets the list separator by precedence level
-getSeparatorByPrecedence :: [Rule] -> [(Integer,String)]
-getSeparatorByPrecedence rules = [ (p, getCons (getRulesFor p)) | p <- precedences ]
-  where
-    precedences = Set.toDescList . Set.fromList $ map precRule rules
-    getRulesFor p = [ r | r <- rules, precRule r == p ]
+sortRulesByPrecedence :: [Rule] -> [(Integer,Rule)]
+sortRulesByPrecedence = List.sortOn (Down . fst) . map (precRule &&& id)
 
+-- | Is the given category a list category parsing also empty lists?
 isEmptyListCat :: CF -> Cat -> Bool
-isEmptyListCat cf c = elem "[]" $ map (wpThing . funRule) $ rulesForCat' cf c
+isEmptyListCat cf = any isNilFun . rulesForCat' cf
 
 isNonterm :: Either Cat String -> Bool
-isNonterm (Left _) = True
-isNonterm (Right _) = False
+isNonterm = Either.isLeft
 
 -- used in Happy to parse lists of form 'C t [C]' in reverse order
 -- applies only if the [] rule has no terminals
@@ -749,10 +738,10 @@ revSepListRule (Rule f c ts internal) = Rule f c (xs : x : sep) internal where
 -- invariant: test in findAllReversibleCats have been performed
 
 findAllReversibleCats :: CF -> [Cat]
-findAllReversibleCats cf = [c | (c,r) <- ruleGroups cf, isRev c r] where
-  isRev c rs = case rs of
-     [r1,r2] | isList c -> if isConsFun (funRule r2)
-                             then tryRev r2 r1
+findAllReversibleCats cf = [c | (c,r) <- ruleGroups cf, isRev c r]
+  where
+  isRev c = \case
+     [r1,r2] | isList c -> if isConsFun (funRule r2) then tryRev r2 r1
                            else isConsFun (funRule r1) && tryRev r1 r2
      _ -> False
   tryRev (Rule f _ ts@(x:_:_) _) r = isEmptyNilRule r &&
@@ -779,6 +768,10 @@ precRule = precCat . valCat
 hasIdentLikeTokens :: CFG g -> Bool
 hasIdentLikeTokens cf = hasIdent cf || or [ not b | TokenReg _ b _ <- cfgPragmas cf ]
 
+-- | Defines or uses the grammar @token@ types or @Ident@?
+hasTextualTokens :: CFG g -> Bool
+hasTextualTokens cf = hasIdent cf || or [ True | TokenReg{} <- cfgPragmas cf ]
+
 -- | Is there a @position token@ declaration in the grammar?
 hasPositionTokens :: CFG g -> Bool
 hasPositionTokens cf = or [ b | TokenReg _ b _ <- cfgPragmas cf ]
