@@ -52,6 +52,7 @@ import BNFC.Par
 import BNFC.CF
 import BNFC.Check.EmptyTypes
 import BNFC.Options
+import BNFC.PrettyPrint
 import BNFC.Regex       (nullable, simpReg)
 import BNFC.TypeChecker
 import BNFC.Utils
@@ -70,7 +71,9 @@ parseCF opts target content = do
                     >>= return . expandRules
                     >>= getCF opts
                     >>= return . markTokenCategories
-  either dieUnlessForce return $ runTypeChecker $ checkDefinitions cf
+
+  -- Construct the typing information in 'define' expressions.
+  cf <- either die return $ runTypeChecker $ checkDefinitions cf
 
   -- Some backends do not allow the grammar name to coincide with
   -- one of the category or constructor names.
@@ -90,13 +93,24 @@ parseCF opts target content = do
           ]
 
   -- Some (most) backends do not support layout.
-  when (hasLayout cf && target `notElem`
+  let (layoutTop, layoutKeywords, _) = layoutPragmas cf
+  let lay = isJust layoutTop || not (null layoutKeywords)
+  when (lay && target `notElem`
     [ TargetHaskell, TargetHaskellGadt, TargetLatex, TargetPygments, TargetCheck ]) $
       dieUnlessForce $ unwords
         [ "ERROR: the grammar uses layout, which is not supported by backend"
         , show target ++ "."
         ]
 
+  -- A grammar that uses layout needs to contain symbols { } ;
+  let symbols = cfgSymbols cf
+      layoutSymbols = concat [ [";"], unless (null layoutKeywords) ["{", "}"] ]
+      missingLayoutSymbols = filter (`notElem` symbols) layoutSymbols
+  when (lay && not (null missingLayoutSymbols)) $
+      dieUnlessForce $ unwords $
+        "ERROR: the grammar uses layout, but does not mention"
+        : map show missingLayoutSymbols
+
   -- Token types that end in a numeral confuse BNFC, because of CoerceCat.
   let userTokenTypes = [ rx | TokenReg rx _ _ <- cfgPragmas cf ]
   case filter (isJust . hasNumericSuffix . wpThing) userTokenTypes of
@@ -147,7 +161,8 @@ parseCF opts target content = do
             ]
 
   -- Warn or fail if the grammar uses names not unique modulo upper/lowercase.
-  case nub $ filter (`notElem` nonUniqueNames) $ filter (not . isDefinedRule) $
+  when False $
+   case nub $ filter (`notElem` nonUniqueNames) $ filter (not . isDefinedRule) $
        concatMap List1.toList $ duplicatesOn (map toLower . wpThing) names of
     [] -> return ()
     ns | target `elem` [ TargetJava ]
@@ -172,10 +187,7 @@ parseCF opts target content = do
         ]
 
   -- Fail if the grammar uses defined constructors which are not actually defined.
-  let definedConstructor = \case
-        FunDef x _ _ -> Just x
-        _ -> Nothing
-  let definedConstructors = Set.fromList $ mapMaybe definedConstructor $ cfgPragmas cf
+  let definedConstructors = Set.fromList $ map defName $ definitions cf
   let undefinedConstructor x = isDefinedRule x && x `Set.notMember` definedConstructors
   case filter undefinedConstructor $ map funRule $ cfgRules cf of
     [] -> return ()
@@ -363,13 +375,13 @@ transDef = \case
         "The delimiters pragma " ++ removedIn290
     Abs.Coercions ident int       -> map Right <$> coercionRules ident int
     Abs.Rules ident strs          -> map Right <$> ebnfRules ident strs
-    Abs.Layout ss                 -> return [ Left $ Layout ss ]
+    Abs.Layout ss                 -> return [ Left $ Layout $ map (,Delimiters ";" "{" "}") ss ]
     Abs.LayoutStop ss             -> return [ Left $ LayoutStop ss]
-    Abs.LayoutTop                 -> return [ Left $ LayoutTop ]
+    Abs.LayoutTop                 -> return [ Left $ LayoutTop ";" ]
     Abs.Function ident xs e       -> do
       f <- transIdent ident
       let xs' = map transArg xs
-      return [ Left $ FunDef f xs' $ transExp xs' e ]
+      return [ Left $ FunDef $ Define f xs' (transExp (map fst xs') e) dummyBase ]
 
 -- | Translate @separator [nonempty] C "s"@.
 --   The position attached to the generated rules is taken from @C@.
@@ -490,8 +502,8 @@ transIdent (Abs.Identifier ((line, col), str)) = do
   file <- asks lbnfFile
   return $ WithPosition (Position file line col) str
 
-transArg :: Abs.Arg -> String
-transArg (Abs.Arg (Abs.Identifier (_pos, x))) = x
+transArg :: Abs.Arg -> (String, Base)
+transArg (Abs.Arg (Abs.Identifier (_pos, x))) = (x, dummyBase)
 
 transExp
   :: [String] -- ^ Arguments of definition (in scope in expression).
@@ -500,17 +512,17 @@ transExp
 transExp xs = loop
   where
   loop = \case
-    Abs.App x es    -> App (transIdent' x) (map loop es)
+    Abs.App x es    -> App (transIdent' x) dummyType (map loop es)
     Abs.Var x       -> let x' = transIdent' x in
-                       if x' `elem` xs then Var x' else App x' []
+                       if x' `elem` xs then Var x' else App x' dummyType []
     Abs.Cons e1 e2  -> cons e1 (loop e2)
     Abs.List es     -> foldr cons nil es
     Abs.LitInt x    -> LitInt x
     Abs.LitDouble x -> LitDouble x
     Abs.LitChar x   -> LitChar x
     Abs.LitString x -> LitString x
-  cons e1 e2 = App "(:)" [loop e1, e2]
-  nil        = App "[]" []
+  cons e1 e2 = App "(:)" dummyType [loop e1, e2]
+  nil        = App "[]"  dummyType []
   transIdent' (Abs.Identifier (_pos, x)) = x
 
 --------------------------------------------------------------------------------
@@ -542,7 +554,7 @@ checkTokens cf
 -- (2) no other digits are used
 
 checkRule :: CF -> Rule -> Maybe String
-checkRule cf (Rule f (WithPosition _ cat) rhs _)
+checkRule cf r@(Rule f (WithPosition _ cat) rhs _)
   | Cat ('@':_) <- cat = Nothing -- Generated by a pragma; it's a trusted category
   | badCoercion = stdFail txtCoercion "Bad coercion in rule"
   | badNil      = stdFail txtNil      "Bad empty list rule"
@@ -550,7 +562,7 @@ checkRule cf (Rule f (WithPosition _ cat) rhs _)
   | badCons     = stdFail txtCons     "Bad list construction rule"
   | badList     = stdFail txtList     "Bad list formation rule"
   | badSpecial  = failure $ "Bad special category rule" +++ s
-  | badTypeName = failure $ "Bad type name" +++ unwords (map show badTypes) +++ "in" +++ s
+  | badTypeName = failure $ "Bad type name" +++ unwords (map prettyShow badTypes) +++ "in" +++ s
   | badFunName  = failure $ "Bad constructor name" +++ fun +++ "in" +++ s
   | badMissing  = failure $ "no production for" +++ unwords missing ++ ", appearing in rule\n    " ++ s
   | otherwise   = Nothing
@@ -559,7 +571,7 @@ checkRule cf (Rule f (WithPosition _ cat) rhs _)
    stdFail txt err = failure $ unlines [ err ++ ":", "  " ++ s, txt ]
 
    fun = wpThing f
-   s  = fun ++ "." +++ show cat +++ "::=" +++ unwords (map (either show show) rhs) -- Todo: consider using the show instance of Rule
+   s  = prettyShow r
    c  = normCat cat                  -- lhs cat without the coercion number
    cs = [normCat c | Left c <- rhs]  -- rhs cats without the coercion numbers
 
@@ -581,9 +593,9 @@ checkRule cf (Rule f (WithPosition _ cat) rhs _)
    badSpecial  = elem c [ Cat x | x <- specialCatsP] && not (isCoercion f)
 
    badMissing  = not (null missing)
-   missing     = filter (`notElem` defineds) [show c | Left c <- rhs]
+   missing     = filter (`notElem` defineds) [catToStr c | Left c <- rhs]
      where
-     defineds = tokenNames cf ++ specialCatsP ++ map (show . valCat) (cfgRules cf)
+     defineds = tokenNames cf ++ specialCatsP ++ map (catToStr . valCat) (cfgRules cf)
 
    badTypeName = not (null badTypes)
    badTypes = filter isBadType $ cat : [c | Left c <- rhs]
