@@ -11,6 +11,7 @@ module BNFC.Options
   , defaultOptions, isDefault, printOptions
   , AlexVersion(..), HappyMode(..), OCamlParser(..), JavaLexerParser(..)
   , RecordPositions(..), TokenText(..)
+  , InPackage
   , removedIn290
   , translateOldOptions
   )
@@ -95,6 +96,9 @@ data JavaLexerParser = JLexCup | JFlexCup | Antlr4
 data RecordPositions = RecordPositions | NoRecordPositions
     deriving (Eq,Show,Ord)
 
+-- | Package name (C++ and Java backends).
+type InPackage = Maybe String
+
 -- | How to represent token content in the Haskell backend?
 
 data TokenText
@@ -112,7 +116,7 @@ data SharedOptions = Options
   , force       :: Bool            -- ^ Ignore errors as much as possible?
   , target      :: Target          -- ^ E.g. @--haskell@.
   , make        :: Maybe String    -- ^ The name of the Makefile to generate or Nothing for no Makefile.
-  , inPackage   :: Maybe String    -- ^ The hierarchical package to put the modules in, or Nothing.
+  , inPackage   :: InPackage       -- ^ The hierarchical package to put the modules in, or Nothing.
   , linenumbers :: RecordPositions -- ^ Add and set line_number field for syntax classes
   --- Haskell specific:
   , inDir         :: Bool        -- ^ Option @-d@.
@@ -336,7 +340,7 @@ specificOptions =
   --         "Use Alex 3 as Haskell lexer tool (default)"
   --   , haskellTargets )
   , ( Option []    ["bytestrings"] (NoArg (\o -> o { tokenText = ByteStringToken }))
-          "Use ByteString in Alex lexer"
+          "Use ByteString in Alex lexer [deprecated, use --text-token]"
     , haskellTargets )
   , ( Option []    ["text-token"] (NoArg (\o -> o { tokenText = TextToken }))
           "Use Text in Alex lexer"
@@ -620,8 +624,8 @@ translateOldOptions = mapM $ \ o -> do
     , ("-alex1"        , "--alex1")
     , ("-alex2"        , "--alex2")
     , ("-alex3"        , "--alex3")
-    , ("-sharestrings" , "--sharestring")
-    , ("-bytestrings"  , "--bytestring")
+    , ("-sharestrings" , "--sharestrings")
+    , ("-bytestrings"  , "--bytestrings")
     , ("-glr"          , "--glr")
     , ("-xml"          , "--xml")
     , ("-xmlt"         , "--xmlt")
